import numpy as np
import math
from numpy import linalg as LG


pi = math.pi

# Our general class will read a Current file and a Voltages file to output a Dirichlet to Neumann map.
# It basis itself on:
# r - radius of body (assumed symmetrical for now) 
# AE -  area of the electrodes
# L - Number of Electrodes
class read_data:
    def __init__(self, str_Object, r, AE, L):
        self.L = L
        self.r = r
        self.AE = AE
        self.Current = np.loadtxt("EIT_Data/"+str_Object+"Current.txt").transpose() # Matrix L x L-1
        self.Voltage = np.loadtxt("EIT_Data/"+str_Object+"Voltage.txt").transpose() # Matrix L x L-1
        self.DNmap = np.zeros((L-1, L-1))
        self.load()
        
    def load(self):
        S=0
        for n in range(self.L-1):
            S = np.sum(self.Voltage[:, n])
            self.Voltage[:, n] = (self.Voltage[:, n]-np.ones(self.L)*S/self.L)/LG.norm(self.Current[:, n], 2)
            self.Current[:, n] = self.Current[:, n]/LG.norm(self.Current[:, n], 2)
            
        for n in range(self.L-1):
            for m in range(self.L-1):
                C = self.Current[:, n]
                V = self.Voltage[:, m]
                self.DNmap[n, m] = np.inner(C, V)
                
        self.DNmap = (self.AE/self.r)*LG.inv(self.DNmap)
        



import numpy as np
import cmath
import math
# Contains Linear Operator and GMRES
import scipy.sparse.linalg as spla
from numpy import linalg as LG
import scipy.linalg as sla
from scipy.fft import fft, ifft
import pyamg
import matplotlib.pyplot as plt


pi = math.pi

class dBar:
    
    def __init__(self, k_grid, Now, Ref, R_z, m_z):
        self.tK = np.zeros(k_grid.k.size, dtype=complex)
        self.Scat_B(Now, Ref, k_grid)
        self.Z = np.zeros(int(pow(2, 2*m_z)), dtype=complex)
        self.load_mesh(1, m_z)
        self.sigma = np.zeros(self.Z.size)
        
        
    def Scat_B(self, Now, Ref, k_grid):
    
        G0 = np.zeros((Now.L, Now.L), dtype=complex)


        dt = (2*pi)/Now.L
        zt = np.exp(1j*np.arange(0, 2*pi, 2*pi/Now.L))

        for l in range(Now.L):
            for ll in range(Now.L):
                if l != ll:
                    G0[l, ll] = -(1/(2*pi))*cmath.log( abs( zt[l] - zt[ll] ) )

        Ez = np.zeros(Now.L, dtype=complex)

        dL = Now.DNmap - Ref.DNmap
        Phi = np.matmul(Now.Current.transpose(),Now.Current)
        M = np.zeros((Now.L-1, Now.L-1), dtype=complex)
        PhidL = np.matmul(Now.Current,dL)

        ind0 = 1e-7

        M = Phi + np.matmul(Now.Current.transpose(),np.matmul(G0, PhidL))

        for l in range(k_grid.k.size):
                if( abs(k_grid.k[l]) < k_grid.R and abs(k_grid.k[l]) > ind0):

                    Ez = np.exp(1j*k_grid.k[l]*zt)

                    psi_b, residuals, rank, s  = np.linalg.lstsq(M, np.matmul(Now.Current.transpose(),Ez), rcond=None)


                    for n in range(Now.L):
                        c = cmath.exp(1j*((k_grid.k[l]*zt[n]).conjugate()))
                        for m in range(Now.L-1):

                            self.tK[l] = self.tK[l] + c*psi_b[m]*PhidL[n, m]

                    self.tK[l] = Now.AE*self.tK[l]/(4*pi*(k_grid.k[l].conjugate()))


    
    def dBar(self, mu, k_grid, zz):
    
        RHS = np.zeros((k_grid.N*k_grid.N), dtype=complex)

        N = len(k_grid.idx)

        for l in range(N):
            RHS[k_grid.idx[l]] = cmath.exp(-2j*( (k_grid.k[k_grid.idx[l]]*zz).real) )* self.tK[k_grid.idx[l]]*complex(mu[l], -mu[l+N])


        F_RHS = fft(RHS, norm="backward")

        F_RHS = np.multiply(F_RHS, k_grid.FG)

        RHS = ifft(F_RHS, norm="backward")


        for l in range(N):
            mu[l] = mu[l] - (k_grid.h*k_grid.h)*RHS[k_grid.idx[l]].real
            mu[l+N] = mu[l+N] - (k_grid.h*k_grid.h)*RHS[k_grid.idx[l]].imag


        return mu
    
    
    def load_mesh(self, R, m):
    
        N = int(pow(2, m))
        h = 2*R/N

        for l in range(N):
            if l%2==0:
                for ll in range(N):
                    self.Z[l*N+ ll] = complex(-R + l*h, -R + ll*h)
            else:
                for ll in range(N):
                    self.Z[l*N+ ll] = complex(-R + l*h, R - h*ll)
                
    
    
    # We are assuming that the data is in EIT_Data/Simulation/...
    def solve(self, k_grid):
        
        N = len(k_grid.idx)
    
        # Define the b and initial solution as the vectors of 1+0.j
        b = np.concatenate((np.ones(N), np.zeros(N)), axis=None)
        mu = np.concatenate((np.ones(N), np.zeros(N)), axis=None)

        
        zz = self.Z[0]

        def Op(mu):
            return self.dBar(mu, k_grid, zz)

        A = spla.LinearOperator((2*N,2*N), matvec=Op)


        for i in range(1, self.Z.size):
                zz = self.Z[i]

                mu, exitcode = pyamg.krylov.gmres(A, b, x0=mu, maxiter=10, orthog='householder')
                #mu = EIT_GMRES(b, mu, 5, Kp, tK, Zz, FG)

                if(abs(zz) <= 1):
                    self.sigma[i] = mu[k_grid.indx]*mu[k_grid.indx]-mu[k_grid.indx+N]*mu[k_grid.indx+N]


    def plot(self):
        
        Z_N = int(math.sqrt(self.Z.size))
        X = np.zeros(Z_N)
        Y = np.zeros(Z_N)
        h = 2/(Z_N)
        for i in range(Z_N):
            X[i] = - 1 + i*h
            Y[i] = - 1 + i*h
            
        sigma_x = np.zeros((Z_N, Z_N))

        for i in range(Z_N):
            if i%2==0:
                for j in range(Z_N):
                    sigma_x[i, j] = self.sigma[j+i*Z_N]
            else:
                for j in range(Z_N):
                    sigma_x[i, j] = self.sigma[(Z_N-1-j)+i*Z_N]

        sigma_x = np.ma.masked_where(sigma_x==0, sigma_x)
        plt.pcolormesh(X, Y, sigma_x, cmap='binary')
        plt.colorbar()   
    



import numpy as np
import cmath
import math
from scipy.fft import fft, ifft


pi = math.pi

class k_grid:
    
    def __init__(self, R, m):
        self.R = R
        self.m = m
        self.s = 2.3*R
        self.h = 2*2.3*R/pow(2, m)
        self.N = pow(2, m)
        self.k = np.zeros(int(self.N*self.N), dtype=complex)
        self.FG = np.zeros(int(self.N*self.N), dtype=complex)
        self.k_gen()
        self.Green_FS()
        self.find()
            
    def k_gen(self):
            
            for l in range(self.N):
                for ll in range(self.N):
                    self.k[l*self.N + ll] = complex(-2.3*self.R + l*self.h, -2.3*self.R + ll*self.h)
                    
    def find(self):
    
        idx = []
        self.indx = -1

        for l in range(self.k.size):
            if abs(self.k[l]) < self.R:
                idx.append(l)
                if(abs(self.k[l]) < 1e-7):
                    self.indx = len(idx)-1

        self.idx = np.array(idx)
    
         
    
    def Green_FS(self):
        eps = self.s/10
        RR = (self.s - eps)/2
        ind0 = 1e-7

        FG = np.zeros(int(self.N*self.N), dtype=complex)

        for l in range(self.k.size):
            if(abs(self.k[l]) < self.s and abs(self.k[l])>ind0 ):

                FG[l] = 1./(pi*self.k[l])

                if(abs(self.k[l]) >= 2*RR):

                    FG[l] = FG[l]*(1. - (abs(self.k[l]) - 2*RR )/eps)


        G =  np.zeros(int(self.N*self.N), dtype=complex)
        G[:] = FG

        ss=0
        ll=0
        N = int(pow(2, self.m))


        for k in range(N):
            for j in range(N):
                ss = int(j + N/2) % N
                ll = int(k + N/2) % N
                FG[j+k*N] = G[ss+ll*N]


        self.FG = fft(FG, norm="backward")



from setuptools import setup, find_packages
import codecs
import os

VERSION = '0.0.10'
DESCRIPTION = 'Dbar Algorithm for EIT'
LONG_DESCRIPTION = 'This package complements the pyEIT package, providing a framework of the Dbar Algorithm for Electrical Impedance Tomography initially implemented by S. Siltanen in Matlab based on the theoretical of A. Nachman'

setup(
    name="py_dbar",
    version=VERSION,
    author="NablaIp",
    author_email = "<ivanpombo.eigen@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type = "text/markdown",
    long_description=LONG_DESCRIPTION,
    packages = find_packages(),
    install_requires = ["numpy", "scipy", "pyamg" ],
    keywords=['python', 'EIT', 'DBar Algorithm'],
    classifiers=[
       'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
    ]

)
